unit amstrad_cpc;

interface
uses sdl,{$IFDEF WINDOWS}windows,{$ENDIF}
     nz80,controls_engine,ay_8910,sysutils,gfx_engine,upd765,cargar_dsk,forms,
     dialogs,rom_engine,misc_functions,main_engine,pal_engine,
     sound_engine,tape_window,file_engine,ppi8255,timer_engine;

const
  cpc_paleta:array[0..31] of dword=(
        $808080,$808080,$00FF80,$FFFF80,
        $000080,$FF0080,$008080,$FF8080,
        $FF0080,$FFFF80,$FFFF00,$FFFFFF,
        $FF0000,$FF00FF,$FF8000,$FF80FF,
        $000080,$00FF80,$00FF00,$00FFFF,
        $000000,$0000FF,$008000,$0080FF,
        $800080,$80FF80,$80FF00,$80FFFF,
        $800000,$8000FF,$808000,$8080FF);
  ram_banks:array[0..7,0..3] of byte=(
        (0,1,2,3),(0,1,2,7),(4,5,6,7),(0,3,2,7),
        (0,4,2,3),(0,5,2,3),(0,6,2,3),(0,7,2,3));

  cpc6128_rom:tipo_roms=(n:'cpc6128.rom';l:$8000;p:0;crc:$9e827fe1);
  cpc464_rom:tipo_roms=(n:'cpc464.rom';l:$8000;p:0;crc:$40852f25);
  cpc646_rom:tipo_roms=(n:'cpc664.rom';l:$8000;p:0;crc:$9ab5a036);
  ams_rom:tipo_roms=(n:'amsdos.rom';l:$4000;p:0;crc:$37f1bb31);
  pantalla_largo=400;

var
    cpc_mem:array[0..10,0..$3fff] of byte;
    mem_marco:array[0..3] of byte;
    current_pen,cpc_video_mode:byte;
    lowrom,highrom:boolean;
    cpcpal:array[0..16] of byte;
    latch_ppi_read_a,latch_ppi_write_a,ay_control:byte;
    crtcreg:array[0..31] of byte;
    current_crtc_reg,keyline:byte;
    keyl_val:array[0..9] of byte;
    rom_selected,galines,galc,sample_beeper,crtc_vsync:byte;
    cont_vsync,long_char:word;
    char_hsync,char_count,char_total:word;
    linea_real,altura_char,vsync_lines,linea_ocurre_vsync:word;
    lineas_total,pixel_visible,altura_borde,lineas_visible:word;
    timer_video:byte;

procedure Cargar_amstrad_CPC;
function iniciar_cpc:boolean;
procedure cpc_main;
procedure cpc_close;
procedure cpc_reset;
//Tape/Disk
function amstrad_tapes:boolean;
procedure amstrad_despues_instruccion(estados_t:word);
function amstrad_loaddisk:boolean;
//Main CPU
function cpc_getbyte(direccion:word):byte;
procedure cpc_putbyte(direccion:word;valor:byte);
function cpc_inbyte(puerto:word):byte;
procedure cpc_outbyte(valor:byte;puerto:word);
function cpc_porta_read:byte;
procedure amstrad_raised_z80;
procedure amstrad_sound_update;
//Video
procedure draw_line;
//PPI 8255
function port_a_read:byte;
function port_b_read:byte;
procedure port_a_write(valor:byte);
procedure port_c_write(valor:byte);

implementation
uses principal,tap_tzx;

procedure Cargar_amstrad_CPC;
begin
llamadas_maquina.iniciar:=iniciar_cpc;
llamadas_maquina.bucle_general:=cpc_main;
llamadas_maquina.reset:=cpc_reset;
llamadas_maquina.fps_max:=50;
llamadas_maquina.velocidad_cpu:=4000000;
llamadas_maquina.cerrar:=cpc_close;
llamadas_maquina.cintas:=amstrad_tapes;
llamadas_maquina.cartuchos:=amstrad_loaddisk;
end;

function iniciar_cpc:boolean;
var
  f:byte;
  colores:tpaleta;
  memoria_temp:array[0..$7fff] of byte;
begin
//Lateral
form1.Panel2.Visible:=true;
//Botones Lateral
form1.BitBtn9.visible:=true;
form1.BitBtn11.visible:=true;
form1.BitBtn10.visible:=true;
form1.BitBtn10.Glyph:=nil;
form1.ImageList2.GetBitmap(3,form1.BitBtn10.Glyph);
form1.BitBtn9.enabled:=true;
form1.BitBtn11.enabled:=true;
if main_vars.tipo_maquina<>7 then form1.BitBtn10.enabled:=true;
iniciar_cpc:=false;
iniciar_audio(false);
p_final[0].x:=pantalla_largo;
p_final[0].y:=312;
p_final[1].x:=pantalla_largo;
p_final[1].y:=312;
iniciar_video;
for f:=0 to 31 do begin
  colores[f].r:=cpc_paleta[f] shr 16;
  colores[f].g:=(cpc_paleta[f] shr 8) and $FF;
  colores[f].b:=cpc_paleta[f] and $FF;
end;
set_pal(colores,32);
main_z80:=cpu_z80.create(4000000,1);
main_z80.change_ram_calls(cpc_getbyte,cpc_putbyte);
main_z80.change_io_calls(cpc_inbyte,cpc_outbyte);
main_z80.change_misc_calls(amstrad_despues_instruccion,amstrad_raised_z80);
//ga_timer:=init_timer(0,4000000/50/312/64,amstrad_hsync,true);
timer_video:=init_timer(0,4000000/312/50,draw_line,true);
sample_beeper:=init_channel;
//El CPC lee el teclado el puerto A del AY, pero el puerto B esta unido al A
//por lo que hay programas que usan el B!!! (Bestial Warrior por ejemplo)
ay8910_0:=ay8910_chip.create(1000000,1);
ay8910_0.change_io_calls(cpc_porta_read,cpc_porta_read,nil,nil);
main_z80.init_sound(amstrad_sound_update);
init_ppi8255(0,port_a_read,port_b_read,nil,port_a_write,nil,port_c_write);
case main_vars.tipo_maquina of
  7:if not(cargar_roms(@memoria_temp[0],@cpc464_rom,'cpc464.zip',1)) then exit;
  8:begin
      if not(cargar_roms(@cpc_mem[10,0],@ams_rom,'cpc664.zip',1)) then exit;
      if not(cargar_roms(@memoria_temp[0],@cpc646_rom,'cpc664.zip',1)) then exit;
  end;
  9:begin
      if not(cargar_roms(@cpc_mem[10,0],@ams_rom,'cpc6128.zip',1)) then exit;
      if not(cargar_roms(@memoria_temp[0],@cpc6128_rom,'cpc6128.zip',1)) then exit;
  end;
end;
copymemory(@cpc_mem[8,0],@memoria_temp[0],$4000);
copymemory(@cpc_mem[9,0],@memoria_temp[$4000],$4000);
cpc_reset;
iniciar_cpc:=true;
end;

procedure cpc_close;
begin
main_z80.free;
close_ppi8255(0);
ay8910_0.free;
clear_disk(0);
clear_disk(1);
close_audio;
close_video;
end;

procedure cpc_reset;
begin
  main_z80.reset;
  ay8910_0.reset;
  reset_ppi8255(0);
  reset_audio;
  if cinta_tzx.cargada then cinta_tzx.play_once:=false;
  cinta_tzx.value:=0;
  ResetFDC;
  //Init GA
  linea_real:=0;
  current_pen:=0;
  cpc_video_mode:=1;
  copymemory(@mem_marco[0],@ram_banks[0,0],4);
  lowrom:=true;
  highrom:=true;
  rom_selected:=9;
  //CRT
  char_count:=0;
  char_hsync:=0;
  cont_vsync:=0;
  linea_real:=0;
  altura_char:=0;
  vsync_lines:=0;
  lineas_total:=0;
  pixel_visible:=0;
  altura_borde:=0;
  lineas_visible:=0;
  current_crtc_reg:=0;
  linea_ocurre_vsync:=0;
  //Keyboard
  fillchar(keyl_val[0],10,$FF);
end;

procedure eventos_cpc;
begin
if event.keyboard then begin
//Line 0
  if keyboard[SDLK_UP] then keyl_val[0]:=(keyl_val[0] and $fe) else keyl_val[0]:=(keyl_val[0] or $1);
  if keyboard[SDLK_RIGHT] then keyl_val[0]:=(keyl_val[0] and $fd) else keyl_val[0]:=(keyl_val[0] or $2);
  if keyboard[SDLK_DOWN] then keyl_val[0]:=(keyl_val[0] and $fb) else keyl_val[0]:=(keyl_val[0] or $4);
  if (keyboard[SDLK_f9] and keyboard[SDLK_RSHIFT])then keyl_val[0]:=(keyl_val[0] and $f7) else keyl_val[0]:=(keyl_val[0] or $8);
  if (keyboard[SDLK_f6] and keyboard[SDLK_RSHIFT])then keyl_val[0]:=(keyl_val[0] and $ef) else keyl_val[0]:=(keyl_val[0] or $10);
  if (keyboard[SDLK_f3] and keyboard[SDLK_RSHIFT])then keyl_val[0]:=(keyl_val[0] and $df) else keyl_val[0]:=(keyl_val[0] or $20);
  if keyboard[SDLK_HOME] then keyl_val[0]:=(keyl_val[0] and $bf) else keyl_val[0]:=(keyl_val[0] or $40);
{F Dot}
//Line 1
  if keyboard[SDLK_LEFT] then keyl_val[1]:=(keyl_val[1] and $fe) else keyl_val[1]:=(keyl_val[1] or 1);
  if keyboard[SDLK_INSERT] then keyl_val[1]:=(keyl_val[1] and $fd) else keyl_val[1]:=(keyl_val[1] or 2);
  if (keyboard[SDLK_f7] and keyboard[SDLK_RSHIFT])then keyl_val[1]:=(keyl_val[1] and $fb) else keyl_val[1]:=(keyl_val[1] or $4);
  if (keyboard[SDLK_f8] and keyboard[SDLK_RSHIFT])then keyl_val[1]:=(keyl_val[1] and $f7) else keyl_val[1]:=(keyl_val[1] or $8);
  if keyboard[SDLK_f5] then begin
    if (keyboard[SDLK_RSHIFT]) then begin
        keyl_val[1]:=(keyl_val[1] and $ef);
    end else begin
        clear_disk(0);
        change_caption(llamadas_maquina.caption);
      end;
  end else begin
        keyl_val[1]:=(keyl_val[1] or $10)
  end;
  if keyboard[SDLK_f1] then begin
    if keyboard[SDLK_RSHIFT] then begin
      keyl_val[1]:=(keyl_val[1] and $df);
    end else begin
      if cinta_tzx.cargada then begin
        if cinta_tzx.play_tape then Form5.fStopCinta(nil)
          else Form5.fPlayCinta(nil);
      end;
    end;
  end else begin
    keyl_val[1]:=(keyl_val[1] or $20);
  end;
  if (keyboard[SDLK_f2] and keyboard[SDLK_RSHIFT])then keyl_val[1]:=(keyl_val[1] and $bf) else keyl_val[1]:=(keyl_val[1] or $40);
  if (keyboard[SDLK_f10] and keyboard[SDLK_RSHIFT])then keyl_val[1]:=(keyl_val[1] and $7f) else keyl_val[1]:=(keyl_val[1] or $80);
//Line 2
  if keyboard[SDLK_DELETE] then keyl_val[2]:=(keyl_val[2] and $fe) else keyl_val[2]:=(keyl_val[2] or 1);
{[}
  if keyboard[SDLK_RETURN] then keyl_val[2]:=(keyl_val[2] and $fb) else keyl_val[2]:=(keyl_val[2] or 4);
{]}
  if (keyboard[SDLK_f4] and keyboard[SDLK_RSHIFT])then keyl_val[2]:=(keyl_val[2] and $ef) else keyl_val[2]:=(keyl_val[2] or $10);
  if keyboard[SDLK_LSHIFT] then keyl_val[2]:=(keyl_val[2] and $df) else keyl_val[2]:=(keyl_val[2] or $20);
{\} if keyboard[SDLK_EQUALS] then keyl_val[2]:=(keyl_val[2] and $bf) else keyl_val[2]:=(keyl_val[2] or $40);
  if keyboard[SDLK_LCTRL] then keyl_val[2]:=(keyl_val[2] and $7f) else keyl_val[2]:=(keyl_val[2] or $80);
//Line 3
{^}
{-} if keyboard[SDLK_SLASH] then keyl_val[3]:=(keyl_val[3] and $fd) else keyl_val[3]:=(keyl_val[3] or 2);
{|} if keyboard[SDLK_LESS] then keyl_val[3]:=(keyl_val[3] and $fb) else keyl_val[3]:=(keyl_val[3] or 4);
  if keyboard[SDLK_p] then keyl_val[3]:=(keyl_val[3] and $f7) else keyl_val[3]:=(keyl_val[3] or 8);
{;} if keyboard[SDLK_QUOTE] then keyl_val[3]:=(keyl_val[3] and $ef) else keyl_val[3]:=(keyl_val[3] or $10);
{:} if keyboard[SDLK_BACKSLASH] then keyl_val[3]:=(keyl_val[3] and $df) else keyl_val[3]:=(keyl_val[3] or $20);
{/} if keyboard[SDLK_MINUS] then keyl_val[3]:=(keyl_val[3] and $bf) else keyl_val[3]:=(keyl_val[3] or $40);
{.} if keyboard[SDLK_PERIOD] then keyl_val[3]:=(keyl_val[3] and $7f) else keyl_val[3]:=(keyl_val[3] or $80);
//Line 4
  if keyboard[SDLK_0] then keyl_val[4]:=(keyl_val[4] and $fe) else keyl_val[4]:=(keyl_val[4] or 1);
  if keyboard[SDLK_9] then keyl_val[4]:=(keyl_val[4] and $fd) else keyl_val[4]:=(keyl_val[4] or 2);
  if keyboard[SDLK_o] then keyl_val[4]:=(keyl_val[4] and $fb) else keyl_val[4]:=(keyl_val[4] or 4);
  if keyboard[SDLK_i] then keyl_val[4]:=(keyl_val[4] and $f7) else keyl_val[4]:=(keyl_val[4] or 8);
  if keyboard[SDLK_l] then keyl_val[4]:=(keyl_val[4] and $ef) else keyl_val[4]:=(keyl_val[4] or $10);
  if keyboard[SDLK_k] then keyl_val[4]:=(keyl_val[4] and $df) else keyl_val[4]:=(keyl_val[4] or $20);
  if keyboard[SDLK_m] then keyl_val[4]:=(keyl_val[4] and $bf) else keyl_val[4]:=(keyl_val[4] or $40);
  if keyboard[SDLK_comma] then keyl_val[4]:=(keyl_val[4] and $7f) else keyl_val[4]:=(keyl_val[4] or $80);
//Line 5
  if keyboard[SDLK_8] then keyl_val[5]:=(keyl_val[5] and $fe) else keyl_val[5]:=(keyl_val[5] or 1);
  if keyboard[SDLK_7] then keyl_val[5]:=(keyl_val[5] and $fd) else keyl_val[5]:=(keyl_val[5] or 2);
  if keyboard[SDLK_u] then keyl_val[5]:=(keyl_val[5] and $fb) else keyl_val[5]:=(keyl_val[5] or 4);
  if keyboard[SDLK_y] then keyl_val[5]:=(keyl_val[5] and $f7) else keyl_val[5]:=(keyl_val[5] or 8);
  if keyboard[SDLK_h] then keyl_val[5]:=(keyl_val[5] and $ef) else keyl_val[5]:=(keyl_val[5] or $10);
  if keyboard[SDLK_j] then keyl_val[5]:=(keyl_val[5] and $df) else keyl_val[5]:=(keyl_val[5] or $20);
  if keyboard[SDLK_n] then keyl_val[5]:=(keyl_val[5] and $bf) else keyl_val[5]:=(keyl_val[5] or $40);
  if keyboard[SDLK_space] then keyl_val[5]:=(keyl_val[5] and $7f) else keyl_val[5]:=(keyl_val[5] or $80);
//Line 6
  if keyboard[SDLK_6] then keyl_val[6]:=(keyl_val[6] and $fe) else keyl_val[6]:=(keyl_val[6] or 1);
  if keyboard[SDLK_5] then keyl_val[6]:=(keyl_val[6] and $fd) else keyl_val[6]:=(keyl_val[6] or 2);
  if keyboard[SDLK_r] then keyl_val[6]:=(keyl_val[6] and $fb) else keyl_val[6]:=(keyl_val[6] or 4);
  if keyboard[SDLK_t] then keyl_val[6]:=(keyl_val[6] and $f7) else keyl_val[6]:=(keyl_val[6] or 8);
  if keyboard[SDLK_g] then keyl_val[6]:=(keyl_val[6] and $ef) else keyl_val[6]:=(keyl_val[6] or $10);
  if keyboard[SDLK_f] then keyl_val[6]:=(keyl_val[6] and $df) else keyl_val[6]:=(keyl_val[6] or $20);
  if keyboard[SDLK_b] then keyl_val[6]:=(keyl_val[6] and $bf) else keyl_val[6]:=(keyl_val[6] or $40);
  if keyboard[SDLK_v] then keyl_val[6]:=(keyl_val[6] and $7f) else keyl_val[6]:=(keyl_val[6] or $80);
//Line 7
  if keyboard[SDLK_4] then keyl_val[7]:=(keyl_val[7] and $fe) else keyl_val[7]:=(keyl_val[7] or 1);
  if keyboard[SDLK_3] then keyl_val[7]:=(keyl_val[7] and $fd) else keyl_val[7]:=(keyl_val[7] or 2);
  if keyboard[SDLK_e] then keyl_val[7]:=(keyl_val[7] and $fb) else keyl_val[7]:=(keyl_val[7] or 4);
  if keyboard[SDLK_w] then keyl_val[7]:=(keyl_val[7] and $f7) else keyl_val[7]:=(keyl_val[7] or 8);
  if keyboard[SDLK_s] then keyl_val[7]:=(keyl_val[7] and $ef) else keyl_val[7]:=(keyl_val[7] or $10);
  if keyboard[SDLK_d] then keyl_val[7]:=(keyl_val[7] and $df) else keyl_val[7]:=(keyl_val[7] or $20);
  if keyboard[SDLK_c] then keyl_val[7]:=(keyl_val[7] and $bf) else keyl_val[7]:=(keyl_val[7] or $40);
  if keyboard[SDLK_x] then keyl_val[7]:=(keyl_val[7] and $7f) else keyl_val[7]:=(keyl_val[7] or $80);
//Line 8
  if keyboard[SDLK_1] then keyl_val[8]:=(keyl_val[8] and $fe) else keyl_val[8]:=(keyl_val[8] or 1);
  if keyboard[SDLK_2] then keyl_val[8]:=(keyl_val[8] and $fd) else keyl_val[8]:=(keyl_val[8] or 2);
  if keyboard[SDLK_escape] then keyl_val[8]:=(keyl_val[8] and $fb) else keyl_val[8]:=(keyl_val[8] or 4);
  if keyboard[SDLK_q] then keyl_val[8]:=(keyl_val[8] and $f7) else keyl_val[8]:=(keyl_val[8] or 8);
  if keyboard[SDLK_tab] then keyl_val[8]:=(keyl_val[8] and $ef) else keyl_val[8]:=(keyl_val[8] or $10);
  if keyboard[SDLK_a] then keyl_val[8]:=(keyl_val[8] and $df) else keyl_val[8]:=(keyl_val[8] or $20);
  if keyboard[SDLK_CAPSLOCK] then keyl_val[8]:=(keyl_val[8] and $bf) else keyl_val[8]:=(keyl_val[8] or $40);
  if keyboard[SDLK_z] then keyl_val[8]:=(keyl_val[8] and $7f) else keyl_val[8]:=(keyl_val[8] or $80);
//Line 9
  //JOY UP,JOY DOWN,JOY LEFT,JOY RIGHT,FIRE1,FIRE2 --> Arcade
  if keyboard[SDLK_BACKSPACE] then keyl_val[9]:=(keyl_val[9] and $7f) else keyl_val[9]:=(keyl_val[9] or $80);
end;
if event.arcade then begin
  //P1
  if arcade_input.up[0] then keyl_val[9]:=(keyl_val[9] and $fe) else keyl_val[9]:=(keyl_val[9] or 1);
  if arcade_input.down[0] then keyl_val[9]:=(keyl_val[9] and $fd) else keyl_val[9]:=(keyl_val[9] or 2);
  if arcade_input.left[0] then keyl_val[9]:=(keyl_val[9] and $fb) else keyl_val[9]:=(keyl_val[9] or 4);
  if arcade_input.right[0] then keyl_val[9]:=(keyl_val[9] and $f7) else keyl_val[9]:=(keyl_val[9] or $8);
  if arcade_input.but0[0] then keyl_val[9]:=(keyl_val[9] and $ef) else keyl_val[9]:=(keyl_val[9] or $10);
  //P2
  if arcade_input.up[1] then keyl_val[6]:=(keyl_val[6] and $fe) else keyl_val[6]:=(keyl_val[6] or 1);
  if arcade_input.down[1] then keyl_val[6]:=(keyl_val[6] and $fd) else keyl_val[6]:=(keyl_val[6] or 2);
  if arcade_input.left[1] then keyl_val[6]:=(keyl_val[6] and $fb) else keyl_val[6]:=(keyl_val[6] or 4);
  if arcade_input.right[1] then keyl_val[6]:=(keyl_val[6] and $f7) else keyl_val[6]:=(keyl_val[6] or 8);
  if arcade_input.but0[1] then keyl_val[6]:=(keyl_val[6] and $ef) else keyl_val[6]:=(keyl_val[6] or $10);
end;
end;

procedure amstrad_ga_exec;inline;
begin
if (galc<>0) then begin
  galc:=galc-1;
  if (galc=0) then begin
    if (galines>=32) then main_z80.pedir_irq:=HOLD_LINE;
    galines:=0;
  end;
end;
if (galines>=52) then begin
    galines:=0;
    main_z80.pedir_irq:=HOLD_LINE;
end;
galines:=galines+1;
end;

procedure actualiza_linea(linea:word);inline;
var
 x,cy:integer;
 addr,addr2:dword;
 marco:byte;
 val,p1,p2,p3,p4,p5,p6,p7,p8:byte;
 ptemp:pword;
 ma,borde:dword;
begin
cy:=linea;//-altura_borde;
if ((linea+altura_borde)>311) then exit;
borde:=(pantalla_largo-pixel_visible) shr 1;
single_line(0,linea+altura_borde,cpcpal[$10],borde,1);
ma:=(crtcreg[13] or (crtcreg[12] shl 8));
addr2:=((ma and $3000) shl 2) or ((ma and $03ff) shl 1);
addr:=addr2 or ((cy mod altura_char) shl 11)+((cy div altura_char)*(crtcreg[1] shl 1));
x:=0;
while (x<char_total) do begin
  if (x div 4)=char_hsync then amstrad_ga_exec;
  if x<pixel_visible then begin
    marco:=addr shr 14;
    val:=cpc_mem[mem_marco[marco and $3],addr and $3fff];
    addr:=addr+1;
    case cpc_video_mode of
      0:begin
          // 1 5 3 7    0 4 2 6
          p1:=((val and 2) shl 2) or ((val and $20) shr 3) or ((val and 8) shr 2) or ((val and $80) shr 7);
          p2:=((val and 1) shl 3) or ((val and $10) shr 2) or ((val and 4) shr 1) or ((val and $40) shr 6);
          ptemp:=punbuf;
          ptemp^:=paleta[cpcpal[p1]];
          inc(ptemp);
          ptemp^:=paleta[cpcpal[p1]];
          inc(ptemp);
          ptemp^:=paleta[cpcpal[p2]];
          inc(ptemp);
          ptemp^:=paleta[cpcpal[p2]];
          putpixel(x+borde,linea+altura_borde,4,punbuf,1);
      end;
      1:begin
          // 3 7      2 6      1 5      0 4
          p1:=((val and $80) shr 7)+((val and $8) shr 2);
          p2:=((val and $40) shr 6)+((val and $4) shr 1);
          p3:=((val and $20) shr 5)+((val and $2) shr 0);
          p4:=((val and $10) shr 4)+((val and 1) shl 1);
          ptemp:=punbuf;
          ptemp^:=paleta[cpcpal[p1]];
          inc(ptemp);
          ptemp^:=paleta[cpcpal[p2]];
          inc(ptemp);
          ptemp^:=paleta[cpcpal[p3]];
          inc(ptemp);
          ptemp^:=paleta[cpcpal[p4]];
          putpixel(x+borde,linea+altura_borde,4,punbuf,1);
        end;
      2:begin
          // 7 6 5 4 3 2 1 0
          p1:=((val and $80) shr 7);
          p2:=((val and $40) shr 6);
          p3:=((val and $20) shr 5);
          p4:=((val and $10) shr 4);
          p5:=((val and $8) shr 3);
          p6:=((val and $4) shr 2);
          p7:=((val and $2) shr 1);
          p8:=((val and $1) shr 0);
          ptemp:=punbuf;
          ptemp^:=(paleta[cpcpal[p1]]+paleta[cpcpal[p2]]) shr 1;
          inc(ptemp);
          ptemp^:=(paleta[cpcpal[p3]]+paleta[cpcpal[p4]]) shr 1;
          inc(ptemp);
          ptemp^:=(paleta[cpcpal[p5]]+paleta[cpcpal[p6]]) shr 1;
          inc(ptemp);
          ptemp^:=(paleta[cpcpal[p7]]+paleta[cpcpal[p8]]) shr 1;
          {inc(ptemp);
          ptemp^:=paleta[cpcpal[p5]];
          inc(ptemp);
          ptemp^:=paleta[cpcpal[p6]];
          inc(ptemp);
          ptemp^:=paleta[cpcpal[p7]];
          inc(ptemp);
          ptemp^:=paleta[cpcpal[p8]];}
          putpixel(x+borde,linea+altura_borde,4,punbuf,1);
      end;
    end;
  end;
  x:=x+4;
end;
single_line(pixel_visible+borde,linea+altura_borde,cpcpal[$10],borde,1);
end;

procedure draw_line;inline;
begin
if linea_real<lineas_visible then actualiza_linea(linea_real) //pintar lo visible
  else begin
    single_line(0,(linea_real+altura_borde) mod 312,cpcpal[$10],pantalla_largo,1);
    amstrad_ga_exec;
  end;
linea_real:=linea_real+1;
if (cont_vsync<>0) then cont_vsync:=cont_vsync-1;
if ((cont_vsync=0) and (crtc_vsync<>0)) then crtc_vsync:=0;
if (linea_real=linea_ocurre_vsync) then begin
 cont_vsync:=vsync_lines;
 galc:=2;
 crtc_vsync:=1;
end;
if (linea_real=312) then begin
  linea_real:=0;
  actualiza_trozo(0,0,pantalla_largo,312,1,0,0,pantalla_largo,312,0);
  video_sync;
end;
end;

procedure cpc_main;
var
  frame:single;
begin
init_controls(false,true,true,false);
frame:=main_z80.tframes;
while EmuStatus=EsRuning do begin
  main_z80.run(trunc(frame));
  frame:=frame+main_z80.tframes-main_z80.contador;
  eventos_cpc;
  evalue_controls;
end;
end;

//Main CPU
function cpc_getbyte(direccion:word):byte;
begin
case direccion of
  0..$3fff:if lowrom then cpc_getbyte:=cpc_mem[8,direccion]
            else cpc_getbyte:=cpc_mem[mem_marco[0],direccion];
  $4000..$7fff:cpc_getbyte:=cpc_mem[mem_marco[1],direccion and $3fff];
  $8000..$bfff:cpc_getbyte:=cpc_mem[mem_marco[2],direccion and $3fff];
  $c000..$ffff:if highrom then cpc_getbyte:=cpc_mem[rom_selected,direccion and $3fff]
                else cpc_getbyte:=cpc_mem[mem_marco[3],direccion and $3fff];
end;
end;

procedure cpc_putbyte(direccion:word;valor:byte);
begin
cpc_mem[mem_marco[direccion shr 14],direccion and $3fff]:=valor;
end;

procedure write_ga(val:byte);inline;
begin
  case (val shr 6) of
          $0:if (val and $10)=0 then current_pen:=val and 15 //Select pen
                else current_pen:=$10; //Border
          $1:cpcpal[current_pen]:=val and 31;    //Change pen colour
          $2:begin   //ROM banking and mode switch
                cpc_video_mode:=val and 3;
                lowrom:=(val and 4)=0;
                highrom:=(val and 8)=0;
                if (val and $10)<>0 then begin
                   galines:=0;
                   main_z80.pedir_irq:=HOLD_LINE;
                end;
            end;
          //C6128 RAM Bank only
          $3:if main_vars.tipo_maquina=9 then copymemory(@mem_marco[0],@ram_banks[(val and 7),0],4);
        end;
end;

{Registros 6845
0 --> Numero de caracteres largo TOTAL (MENOS UNO)
      En el CPC 63
1 --> El ancho en caracteres VISIBLES de una fila.
2 --> Caracter en el que se produce el HSYNC
3 --> 4bits lo: Numero de caracteres que tarda el HSYNC
      4bits hi: Numero de caracteres que tarda el VSYNC
4 --> La altura en caracteres TOTAL de la pantalla +1
      En el CPC 38+1
5 --> Numero de lineas que faltan para completar la pantalla cuando termina un frame
      en el CPC 0
6 --> La altura en caracteres VISIBLES de la pantalla.
7 --> Numero de caracter en el que tiene que ocurrir el VSYNC
9 --> Numero de lineas por caracter (menos una), cuantas lineas necesito para dibujar un caracter
      Fijo a 7+1

Altura TOTAL= R4*(R9+1) = 39*8 = 312 lineas
Ancho TOTAL = (R0+1)*8 = 64*8 = 512 pixels}


procedure write_crtc(port:word;val:byte);inline;
begin
  case ((port and $300) shr 8) of
    $00:current_crtc_reg:=val and 31;
    $01:begin
          crtcreg[current_crtc_reg]:=val;
          //altura char por linea
          altura_char:=(crtcreg[9] and $1f)+1;
          //chars totales
          char_total:=(crtcreg[0]+1)*8;
          //char en el que ocurre el hsync
          char_hsync:=crtcreg[2];
          //Total pixels por linea
          pixel_visible:=crtcreg[1]*8;
          //Lineas de vsync+completar frame
          if (crtcreg[3] shr 4)=0 then vsync_lines:=16+(crtcreg[5] and $1f)
            else vsync_lines:=(crtcreg[3] shr 4)+(crtcreg[5] and $1f);
          //lineas visibles
          lineas_visible:=(crtcreg[6] and $7f)*altura_char;
          //lineas totales del video incluyendo las visibles
          lineas_total:=((crtcreg[4] and $7f)+1)*altura_char;
          if lineas_visible>lineas_total then altura_borde:=0
            else altura_borde:=(lineas_total-lineas_visible) shr 1;
          //Linea donde ocurre el vsync
          linea_ocurre_vsync:=(crtcreg[7] and $7f)*altura_char;
          long_char:=trunc(4000000/50/lineas_total/(crtcreg[0]+1));
          //timer[ga_timer].time_final:=4000000/50/lineas_total/(char_total/2);
        end;
  end;
end;

procedure cpc_outbyte(valor:byte;puerto:word);
begin
if (puerto and $C000)=$4000 then write_ga(valor);
if (puerto and $4000)=0 then write_crtc(puerto,valor);
if ((puerto and $2000)=0) then begin
    case valor of
      0:rom_selected:=9;
      7:rom_selected:=10;
    end;
end;
if (puerto and $0800)=0 then ppi8255_w(0,(puerto and $300) shr 8,valor);
if (puerto and $0581)=$101 then WriteFDCData(valor);
end;

function read_crtc(port:word):byte;inline;
var
  res:byte;
begin
res:=$ff;
if ((port and $300) shr 8)=0 then
  if ((current_crtc_reg>11) and (current_crtc_reg<18)) then res:=crtcreg[current_crtc_reg];
read_crtc:=res;
end;

function cpc_inbyte(puerto:word):byte;
var
  res:byte;
begin
res:=$FF;
if (puerto and $4000)=0 then res:=read_crtc(puerto);
if (puerto and $0800)=0 then res:=ppi8255_r(0,(puerto and $300) shr 8);
if (puerto and $0580)=$100 then
  case (puerto and $101) of
    $100:res:=ReadFDCStatus;
    $101:res:=ReadFDCData;
  end;
cpc_inbyte:=res;
end;

procedure amstrad_raised_z80;
begin
  galines:=galines and $df;
end;

//PPI 8255
procedure update_ay;inline;
begin
case ay_control of
  1:latch_ppi_read_a:=ay8910_0.read;
  2:ay8910_0.write(latch_ppi_write_a);
  3:ay8910_0.control(latch_ppi_write_a);
end;
end;

function port_a_read:byte;
begin
  update_ay;
  port_a_read:=latch_ppi_read_a;
end;

//El valor $7e es fijo para indicar:
// bit 0 vsync
// bits 1-3 el fabricante
// bit 4 refresco pantalla (1-50hz 0-60hz)
// bit 5 expansion port
// bit 6 print ready (1-not ready 0-ready)
// bit 7 cassete data
function port_b_read:byte;
begin
port_b_read:=$7e or (cinta_tzx.value shl 1) or crtc_vsync;
end;

procedure port_a_write(valor:byte);
begin
  latch_ppi_write_a:=valor;
  update_ay;
end;

procedure port_c_write(valor:byte);
begin
ay_control:=((valor and $c0) shr 6);
update_ay;
keyline:=valor and $f;
end;

//AY-8910
function cpc_porta_read:byte;
begin
  cpc_porta_read:=keyl_val[keyline];
end;

//Sound
procedure amstrad_sound_update;
begin
tsample[sample_beeper,sound_status.posicion_sonido]:=cinta_tzx.value*$20;
ay8910_0.update;
end;

//Tape/Disk
procedure amstrad_despues_instruccion(estados_t:word);
begin
if cinta_tzx.cargada then begin
  if cinta_tzx.play_tape then begin
        cinta_tzx.estados:=cinta_tzx.estados+estados_t;
        play_cinta_tzx;
  end;
end;
end;

function amstrad_tapes:boolean;
var
  datos:pbyte;
  file_size,crc:integer;
  nombre_zip,nombre_file,extension:string;
  resultado:boolean;
begin
  amstrad_tapes:=false;
  form1.OpenDialog1.Filter:='CPC Tape (*.cdt;*.csw;*.wav;*.zip)|*.cdt;*.csw;*.wav;*.zip';
  form1.opendialog1.InitialDir:=directory.amstrad_tap;
  if form1.OpenDialog1.execute then begin
    extension:=extension_fichero(form1.OpenDialog1.FileName);
    nombre_zip:=form1.OpenDialog1.FileName;
    if extension='ZIP' then begin
         if not(search_file_from_zip(nombre_zip,'*.cdt',nombre_file,file_size,crc,false)) then
            if not(search_file_from_zip(nombre_zip,'*.csw',nombre_file,file_size,crc,false)) then
              if not(search_file_from_zip(nombre_zip,'*.wav',nombre_file,file_size,crc,false)) then exit;
         getmem(datos,file_size);
         if not(load_file_from_zip(nombre_zip,nombre_file,datos,file_size,crc,true)) then begin
            freemem(datos);
            exit;
         end;
    end else begin
      if not(read_file_size(nombre_zip,file_size)) then exit;
      getmem(datos,file_size);
      if not(read_file(nombre_zip,datos,file_size)) then exit;
      nombre_file:=extractfilename(nombre_zip);
    end;
    extension:=extension_fichero(nombre_file);
    if extension='CDT' then resultado:=abrir_tzx(datos,file_size);
    if extension='CSW' then resultado:=abrir_csw(datos,file_size);
    if extension='WAV' then resultado:=abrir_wav(datos,file_size);
    if resultado then begin
        form5.edit1.Text:=nombre_file;
        form5.show;
        form5.BitBtn1.Enabled:=true;
        form5.BitBtn2.Enabled:=false;
        cinta_tzx.play_tape:=false;
        amstrad_tapes:=true;
    end;
    freemem(datos);
    directory.amstrad_tap:=extractfiledir(nombre_zip)+main_vars.cadena_dir;
  end;
end;

function amstrad_loaddisk:boolean;
begin
form6.show;
while form6.Showing do application.ProcessMessages;
amstrad_loaddisk:=true;
end;


end.
