unit dac;

interface
uses sound_engine;

type
  dac_chip=class(snd_chip_class)
      constructor Create;
      procedure Free;
      destructor Destroy;
    public
      procedure update;
      procedure reset;
      procedure signed_data_w(data:byte);
      procedure data_w(data:byte);
      function save_snapshot(data:pbyte):word;
      procedure load_snapshot(data:pbyte);
    private
      output:integer;
	    UnsignedVolTable:array[0..256-1] of word;
	    SignedVolTable:array[0..256-1] of smallint;
  end;
var
  dac_0,dac_1:dac_chip;

implementation

constructor dac_chip.Create;
var
  i:byte;
begin
  // build volume table (linear) */
	for i:=0 to 255 do begin
		self.UnsignedVolTable[i]:=i*$101 shr 1;	// range      0..32767 */
		self.SignedVolTable[i]:=i*$101-$8000;	// range -32768..32767 */
	end;
  self.tsample_num:=init_channel;
  self.reset;
end;

destructor dac_chip.Destroy;
begin
end;

procedure dac_chip.Free;
begin
end;

function dac_chip.save_snapshot(data:pbyte):word;
var
  ptemp:pbyte;
begin
ptemp:=data;
ptemp^:=self.output and $ff;inc(ptemp);
ptemp^:=(self.output shr 8) and $ff;inc(ptemp);
ptemp^:=(self.output shr 16) and $ff;inc(ptemp);
ptemp^:=(self.output shr 24) and $ff;
save_snapshot:=4;
end;

procedure dac_chip.load_snapshot(data:pbyte);
var
  ptemp:pbyte;
begin
  ptemp:=data;
  self.output:=ptemp^;inc(ptemp);
  self.output:=self.output or (ptemp^ shl 8);inc(ptemp);
  self.output:=self.output or (ptemp^ shl 16);inc(ptemp);
  self.output:=self.output or (ptemp^ shl 24);
end;

procedure dac_chip.update;
begin
  tsample[self.tsample_num,sound_status.posicion_sonido]:=self.output;
  if sound_status.stereo then tsample[self.tsample_num,sound_status.posicion_sonido+1]:=self.output;
end;

procedure dac_chip.data_w(data:byte);
begin
	self.output:=self.UnsignedVolTable[data];
end;

procedure dac_chip.signed_data_w(data:byte);
begin
	self.output:=self.SignedVolTable[data];
end;

procedure dac_chip.reset;
begin
  self.output:=0;
end;

end.
