unit msm5205;

interface
uses math,timer_engine,sound_engine;

const
  index_shift:array[0..7] of integer=(-1, -1, -1, -1, 2, 4, 6, 8);
  MSM5205_S96_3B=0;     // prsicaler 1/96(4KHz) , data 3bit */
  MSM5205_S48_3B=1;     // prsicaler 1/48(8KHz) , data 3bit */
  MSM5205_S64_3B=2;     // prsicaler 1/64(6KHz) , data 3bit */
  MSM5205_SEX_3B=3;     // VCLK slave mode      , data 3bit */
  MSM5205_S96_4B=4;     // prsicaler 1/96(4KHz) , data 4bit */
  MSM5205_S48_4B=5;     // prsicaler 1/48(8KHz) , data 4bit */
  MSM5205_S64_4B=6;     // prsicaler 1/64(6KHz) , data 4bit */
  MSM5205_SEX_4B=7;     // VCLK slave mode      , data 4bit */

type
  MSM5205_tipo=record
    index:integer;
	  clock:integer;				// clock rate */
	  //emu_timer *timer;        // VCLK callback timer          */
	  data:integer;               // next adpcm data              */
	  vclk:integer;               // vclk signal (external mode)  */
	  reset:integer;              // reset pin signal             */
	  prescaler:byte;          // prescaler selector S1 and S2 */
	  bitwidth:byte;           // bit width selector -3B/4B    */
	  signal:integer;             // current ADPCM signal         */
	  step:integer;               // current ADPCM step           */
	  diff_lookup:array[0..(49*16)-1] of integer;
    select:byte;
    snd_cpu_clock:double;
    timer,tsample,amp:byte;
    external_call:procedure;
  end;
  pMSM5205=^MSM5205_tipo;
var
  MSM_5205:array[0..1] of pMSM5205;

procedure msm5205_init(num:byte;clock:integer;num_cpu,select,amp:byte;snd_cpu_clock:double;snd_timer_call:exec_type);
procedure msm5205_close(num:byte);
procedure msm5205_Reset(num:integer);
procedure MSM5205_reset_w(num,reset:byte);
procedure MSM5205_data_w(num,data:byte);
procedure msm5205_internal_update_0;
procedure msm5205_internal_update_1;
procedure msm5205_final_update_0;
procedure msm5205_final_update_1;
procedure msm5205_vclk_w(num,vclk:byte);

implementation

procedure ComputeTables(num:byte);inline;
// nibble to bit map */
const
  nbl2bit:array[0..15,0..3] of integer= (
		( 1, 0, 0, 0), ( 1, 0, 0, 1), ( 1, 0, 1, 0), ( 1, 0, 1, 1),
		( 1, 1, 0, 0), ( 1, 1, 0, 1), ( 1, 1, 1, 0), ( 1, 1, 1, 1),
		(-1, 0, 0, 0), (-1, 0, 0, 1), (-1, 0, 1, 0), (-1, 0, 1, 1),
		(-1, 1, 0, 0), (-1, 1, 0, 1), (-1, 1, 1, 0), (-1, 1, 1, 1));
var
	step,nib,stepval:integer;
  voice:pmsm5205;
begin
  voice:=msm_5205[num];
	// loop over all possible steps */
	for step:=0 to 48 do begin
		// compute the step value */
		stepval:= floor(16.0 * power (11.0 / 10.0,step));
		// loop over all nibbles and compute the difference */
		for nib:=0 to 15 do begin
			voice.diff_lookup[step*16 + nib]:= nbl2bit[nib][0] *
				(stepval   * nbl2bit[nib][1] +
				 stepval div 2 * nbl2bit[nib][2] +
				 stepval div 4 * nbl2bit[nib][3] +
				 stepval div 8);
		end;
	end;
end;

procedure MSM5205_playmode_w(num,select:integer);
const
  prescaler_table:array[0..3] of integer=(96,48,64,0);
var
  voice:pmsm5205;
  prescaler,bitwidth:byte;
begin
  voice:=msm_5205[num];
	prescaler:=prescaler_table[select and 3];
	if (select and 4)<>0 then bitwidth:=4
    else bitwidth:=3;
	if (voice.prescaler<>prescaler) then begin
		voice.prescaler:=prescaler;
		// timer set */
    if (prescaler<>0) then begin
  			timer[voice.timer].time_final:=voice.snd_cpu_clock/(voice.clock/prescaler);
        timer[voice.timer].enabled:=true;
  		end else timer[voice.timer].enabled:=false;
  end;
	if (voice.bitwidth<>bitwidth) then voice.bitwidth:=bitwidth;
end;

procedure msm5205_reset(num:integer);
var
  voice:pMSM5205;
begin
	voice:=MSM_5205[num];
	// initialize work */
	voice.data:=0;
	voice.vclk:=0;
	voice.reset:=0;
	voice.signal:=0;
	voice.step:=0;
	// timer and bitwidth set */
	MSM5205_playmode_w(num,voice.select);
end;

procedure msm5205_close(num:byte);
begin
if msm_5205[num]<>nil then begin
  freemem(msm_5205[num]);
  msm_5205[num]:=nil;
end;
end;

procedure msm5205_init(num:byte;clock:integer;num_cpu,select,amp:byte;snd_cpu_clock:double;snd_timer_call:exec_type);
var
  voice:pmsm5205;
begin
	if msm_5205[num]=nil then begin
    getmem(msm_5205[num],sizeof(msm5205_tipo));
    fillchar(msm_5205[num]^,sizeof(msm5205_tipo),0);
  end;
  voice:=msm_5205[num];
  //voice.amp:=amp;
  voice.prescaler:=$ff;
	// save a global pointer to our interface */
	voice.index:=num;
  voice.amp:=amp;
	voice.clock:=clock;
  voice.select:=select;
  voice.snd_cpu_clock:=snd_cpu_clock;
  voice.tsample:=init_channel;
  voice.external_call:=snd_timer_call;
  case num of
    0:begin
        voice.timer:=init_timer(num_cpu,1,msm5205_internal_update_0,false);
        init_timer(num_cpu,snd_cpu_clock/44100,msm5205_final_update_0,true);
      end;
    1:begin
        voice.timer:=init_timer(num_cpu,1,msm5205_internal_update_1,false);
        init_timer(num_cpu,snd_cpu_clock/44100,msm5205_final_update_1,true);
      end;
  end;
	// compute the difference tables */
	ComputeTables(num);
	// initialize */
	msm5205_reset(num);
end;

procedure msm5205_update(num:byte);
var
  voice:pmsm5205;
begin
  voice:=msm_5205[num];
  tsample[voice.tsample,sound_status.posicion_sonido]:=(voice.signal shl 4)*voice.amp;
  if sound_status.stereo then tsample[voice.tsample,sound_status.posicion_sonido+1]:=(voice.signal shl 4)*voice.amp;
end;

procedure msm5205_stream_update(num:byte);
var
  voice:pmsm5205;
begin
  voice:=msm_5205[num];
  if addr(voice.external_call)<>nil then voice.external_call;
	// reset check at last hieddge of VCLK */
	if (voice.reset<>0) then begin
		voice.signal:=0;
		voice.step:=0;
	end else begin
		// update signal */
		// !! MSM5205 has internal 12bit decoding, signal width is 0 to 8191 !! */
		voice.signal:=voice.signal+voice.diff_lookup[voice.step*16+(voice.data and 15)];
		if (voice.signal>2047) then voice.signal:=2047
     else if (voice.signal<-2048) then voice.signal:=-2048;
		voice.step:=voice.step+index_shift[voice.data and 7];
		if (voice.step>48) then voice.step:=48
		  else if (voice.step<0) then voice.step:=0;
	end;
end;

procedure MSM5205_data_w(num,data:byte);
var
  voice:pmsm5205;
begin
  voice:=msm_5205[num];
	if (voice.bitwidth=4) then voice.data:=data and $0f
    else voice.data:=(data and $07) shl 1; // unknown */
end;

procedure MSM5205_reset_w(num,reset:byte);
begin
  msm_5205[num].reset:=reset;
end;

procedure msm5205_internal_update_0;
begin
  msm5205_stream_update(0);
  //Slave!!
  if msm_5205[1]<>nil then
   if msm_5205[1].prescaler=0 then msm5205_stream_update(1);
end;

procedure msm5205_internal_update_1;
begin
  msm5205_stream_update(1);
end;

procedure msm5205_final_update_0;
begin
  msm5205_update(0);
end;

procedure msm5205_final_update_1;
begin
  msm5205_update(1);
end;

procedure msm5205_vclk_w(num,vclk:byte);
begin
  if (msm_5205[num].vclk<>vclk) then begin
			msm_5205[num].vclk:=vclk;
			if (vclk=0) then begin
        case num of
          0:msm5205_stream_update(0);
          1:msm5205_stream_update(1);
        end;
      end;
  end;
end;

end.
